##########RNA Modification Calling from meRIP-seq
##########Developed based on PEA tool (https://hub.docker.com/r/malab/pea, https://www.ncbi.nlm.nih.gov/pubmed/29850798)

.runMACS2 <- function(IPBAM, inputBAM, broad = FALSE, 
                     expName = NULL, paired = F,MACS2Dir, ...){
  
  tt <- list(...)
  if(length(tt) != 0){
    callPara <- tt[[1]]
  }else{
    callPara <- NULL
  }

  if(is.null(expName)){
      expName <- "example"
    }
    
    if(!broad){
      if(!paired){
        macsCommand <- paste0(MACS2Dir, " callpeak -t ", IPBAM, " -c ", inputBAM, 
                              " -f BAM --nomodel -n ", expName, " ", callPara)
      }else{
        macsCommand <- paste0(MACS2Dir, " callpeak -t ", IPBAM, " -c ", inputBAM, 
                              " -f BAMPE --nomodel -n ", expName, " ", callPara)
      }
      
    }else{
      
      if(!paired){
        macsCommand <- paste0(MACS2Dir, " callpeak -t ", IPBAM, " -c ", inputBAM, 
                              " --broad ", callPara)
      }else{
        macsCommand <- paste0(MACS2Dir, " callpeak -t ", IPBAM, " -c ", inputBAM, 
                              " -f BAMPE --broad ", callPara)
      }
      
    }
    
    system(command = macsCommand)
  
    peakDir <- paste0(expName, "_peaks.narrowPeak")
    peaks <- as.matrix(read.table(file = peakDir, sep = "\t", header = F, quote = ""))
  
    resList <- list(macsCommand = macsCommand, peaks = peaks)
    resList
}



.bam2bed <- function(BAM){
  
  BAMFile <- unlist(strsplit(x = BAM, split = "/", fixed = T))
  BAMFile <- BAMFile[length(BAMFile)]
  BAMName <- unlist(strsplit(x = BAMFile, split = ".", fixed = T))[1]
  bedCommand <- paste0("bedtools bamtobed -i ", BAM, " > ", BAMName, ".bed")
  system(bedCommand)
  resBed <- paste0(BAMName, ".bed")
  resBed
}

extractCov <- function(BAM, refGenome, method = "Rsamtools"){
  
  if(length(method) > 1){
    method <- method[1]
  }
  
  if(method == "bedtools"){
    command <- paste0("genomeCoverageBed -split -dz -ibam ", BAM, " -g ", 
                      refGenome, " > ", paste0(BAM, ".dz"))
    system(command = command)
  }else{
    refGenome <- read.fasta(file = refGenome, as.string = T)
    Chr <- names(refGenome)
    resMat <- NULL
    for(i in 1:length(Chr)){
      seqID <- Chr[i]
      resPos <- 1:(nchar(refGenome[[i]]))
      USiteIRange <- IRanges(resPos, resPos)
      names(USiteIRange) <- seqID
      tmpList <- list()
      tmpList[[seqID]] <- USiteIRange
      which <- RangesList(unlist(tmpList))
      p1 <- ScanBamParam(which = which, what = scanBamWhat())
      res1 <- scanBam(BAM, param = p1)
      number <- lapply(res1, function(x) nrow(as.data.frame(x)))
      tmpMat <- data.frame(Chr = rep(seqID, length(number)), Position = 1:(nchar(refGenome[[i]])),
                           number = unlist(number))
      resMat <- rbind(resMat, res1)
    }
    write.table(resMat, file = paste0(BAM, ".dz"), sep = "\t",
                quote = F, row.names = F, col.names = F)
  }
  res <- paste0(BAM, ".dz")
  res
}


.findSeed <- function(index){
  res1 <- NULL
  res2 <- NULL
  index <- c(index, 0)
  for(i in 1:(length(index)-1)){
    
    if(i >= 2){
      if(((index[i+1] - index[i]) == 1) & ((index[i] - index[i-1])>=2)){
        res1 <- c(res1, i)
      }
      
      if(((index[i] - index[i-1]) == 1) & ((index[i] - index[i+1])!=-1)){
        res2 <- c(res2, i)
      }
    }else{
      if((index[i] - index[i+1]) == -1){
        res1 <- c(res1, i)
      }
    }
    
    
  }
  resMat <- cbind(res1, res2)
  resMat
}

.intervalCov <- function(dzFile){
  
  baseCov <- read.table(file = dzFile, sep = "\t", header = F, stringsAsFactors = F)
  winDow <- rep(NA, nrow(baseCov))
  baseCov <- cbind(baseCov, winDow)
  colnames(baseCov) <- c("Chr", "Position", "readsNumber", "window")
  baseCov$window <- ceiling(baseCov$Position/25)
  
  Chr <- unique(baseCov$Chr)
  resList <- vector("list", length = length(Chr))
  names(resList) <- Chr
  for(i in 1:length(Chr)){
    curChr <- Chr[i]
    curMat <- subset(x = baseCov, baseCov$Chr == curChr)
    curWave <- rep(NA, nrow(curMat))
    curMat <- cbind(curMat, curWave)
    curMat$curWave <- ave(curMat$readsNumber, curMat$window, FUN=mean)
    curMat <- subset(curMat, !duplicated(curMat$window))
    curMat <- na.omit(curMat)
    resList[[curChr]] <- curMat
  }
  resList
}

#' @export
SlidingWindow <- function(input, RIP, refGenome,mappedInput = NULL, 
                          mappedRIP = NULL, level = 0.05, 
                          ratio = -2, ...){
  
  input.dz <- extractCov(BAM = input, refGenome = refGenome, ...)
  RIP.dz <- extractCov(BAM = RIP,refGenome = refGenome, ...)
  input <- .intervalCov(dzFile = input.dz)
  RIP <- .intervalCov(dzFile = RIP.dz)
  
  if(!all(names(input) == names(RIP))){
    cat("Note: The chromosomes in the input and RIP are not consistent!\n", 
        "the interactions will be used!")
    interNames <- intersect(names(input), names(RIP))
    input <- input[interNames]
    RIP <- RIP[interNames]
  }
  
  resList <- vector("list", length = length(RIP))
  names(resList) <- names(RIP)
  #i=1
  for(i in 1:length(input)){
    curMat <- merge(input[[i]], RIP[[i]], by = 'window', all=TRUE)
    curMat$curWave.x[is.na(curMat$curWave.x)] <- 0
    curMat$curWave.y[is.na(curMat$curWave.y)] <- 0
    curMat[, 'windowave.input'] <- round(curMat$curWave.x)
    curMat[, 'windowave.RIP'] <- round(curMat$curWave.y)
    curMat <- curMat[curMat$windowave.RIP != 0, ]
    curPvalue <- rep(NA, nrow(curMat))
    curRatio <- rep(NA, nrow(curMat))
    curFDR <- rep(NA, nrow(curMat))
    curMat <- cbind(curMat, curPvalue, curRatio, curFDR)
    cat("Chromosome: ", names(input)[i], "\n")
    pb <- txtProgressBar(min = 0, max = nrow(curMat), style = 3,width = 75)
    for(j in 1:nrow(curMat)){
      testMat <- matrix(c(curMat$windowave.input[j],
                          mappedInput,
                          curMat$windowave.RIP[j],
                          mappedRIP), nrow = 2, ncol = 2)
      curMat$curPvalue[j] <- fisher.test(x = testMat)$p
      curMat$curRatio[j] <- log2(((curMat$windowave.input[j] + 1)*mappedRIP)/((curMat$windowave.RIP[j] + 1)*mappedInput))
      setTxtProgressBar(pb, j)
    }
    close(pb)
    curMat$curFDR <- p.adjust(curMat$curPvalue, "fdr")
    resList[[i]] <- curMat
  }
  resMat <- do.call(what = rbind, args = resList)
  sig <- rep(NA, nrow(resMat))
  
  for(i in 1:(nrow(resMat)-3)){
    if(resMat$curRatio[i]< ratio & resMat$curRatio[i+1] < ratio & 
       resMat$curRatio[i+2] < ratio & resMat$curRatio[i+3] < ratio & 
       resMat$curFDR[i] < level & resMat$curFDR[i+1] < level & 
       resMat$curFDR[i+2] < level & resMat$curFDR[i+3] < level ){
      sig[i] <- 1
    }
    else {
      sig[i] <- 0
    }
  }
  resMat <- cbind(resMat, sig)
  resMat <- subset(resMat, resMat$sig == 1)
  resPeaks <- NULL
  Chr <- na.omit(unique(resMat[,2]))
  for(i in 1:length(Chr)){
    curPeaks <- subset(resMat, resMat$Chr.x == Chr[i])
    curSeed <- .findSeed(index = curPeaks$window)
    if(!is.null(dim(curSeed))){
      tmpPeaks <- matrix(NA, nrow = nrow(curSeed), ncol = 3)
      tmpPeaks[,1] <- Chr[i]
      for(j in 1:nrow(curSeed)){
        curSeed1 <- curSeed[j,1]
        curSeed2 <- curSeed[j,2]
        curStart <- curPeaks[which(curPeaks$window == curPeaks$window[curSeed1]), 3]
        curEnd <- curPeaks[which(curPeaks$window == curPeaks$window[curSeed2]), 3]
        tmpPeaks[j, 2:3] <- c(curStart, curEnd)
      }
    }
    
    resPeaks <- rbind(resPeaks, tmpPeaks)
  }
  resLen <- as.numeric(resPeaks[,3]) - as.numeric(resPeaks[,2])
  resPeaks <- resPeaks[which(resLen >= 100), ]
  resPeaks
}


peakCalling <- function(IPBAM, inputBAM, GTF, expName = NULL,
                        method = c("SlidingWindow","exomePeak", "MeTPeak", 
                                   "macs2", "BayesPeak"), 
                        paired = F, ...){
  if(length(method) != 1){
    method <- method[1]
    cat("Note: multiple peak calling methods are provided, the first one
        will be used!")
  }
  
  if(is.null(expName)){
    expName <- "example"
  }
  
  if(method == "exomePeak"){
    cat("Start peak calling using exomePeak...", "\n")
    results <- exomepeak(GENE_ANNO_GTF = GTF, IP_BAM = IPBAM,
                         INPUT_BAM = inputBAM, ...)
    peaks <- read.table(paste0(getwd(), '/exomePeak_output/peak.bed'), 
                        sep = "\t", quote = "", stringsAsFactors = F)
    #peaks <- results$all_peaks
    
  }else if(method == "BayesPeak"){
    cat("Start peak calling using BayesPeak...", "\n")
    IPBed <- .bam2bed(BAM = IPBAM)
    inputBed <- .bam2bed(BAM = inputBAM)
    res <- bayespeak(treatment = IPBed, control = inputBed, ...)
    peaks <- res$peaks
  }else if(method == "MACS2"){
    cat("Start peak calling using MACS2...", "\n")
    MACS2Dir <- system("which macs2", intern = TRUE)
    resList <- .runMACS2(IPBAM = IPBAM, inputBAM = inputBAM, 
                         paired = paired, MACS2Dir = MACS2Dir, ...)
    peaks <- resList$peaks
   
  }else if(method == "MeTPeak"){
    cat("Start peak calling using MeTPeak...", "\n")
    results <- metpeak(GENE_ANNO_GTF = GTF,IP_BAM = IPBAM, INPUT_BAM = inputBAM,
                       EXPERIMENT_NAME = expName, ...)
    resDir <- paste0(getwd(), "/", expName, "/peak.bed")
    peaks <- as.matrix(read.table(file = resDir, sep = "\t", quote = "", header = F))
  }else{
    cat("Start peak calling using Fisher exact test-based sliding window...", "\n")
    peaks <- SlidingWindow(input = inputBAM, RIP = IPBAM, refGenome=refGenome, ...)
  }
  resList <- list(peaks = peaks, method = method)
}



.sub <- function(number, resPos, res1){
  curNumber <- as.numeric(resPos[number])
  mapped <- as.data.frame(res1[[number]])
  if(nrow(mapped) == 0){
    ratio <- 0
  }else{
    ratio <- length(which(mapped[,5] == curNumber))/nrow(mapped)
  }
}


#' @export
ModCalling <- function(cpus = 1, IPBAM = NULL, inputBAM = NULL, 
                       GTF = NULL,  paired = FALSE, ...){
    resMat <- peakCalling(IPBAM = IPBAM, inputBAM = inputBAM, GTF = GTF, ...)
    resMat <- resMat$peaks
    resMat
}

#main
if (!"optparse" %in% rownames(installed.packages())) {
	stop(paste0("\n\noptparse package is not available.\nPlease install it (https://cran.r-project.org/web/packages/optparse/index.html) in R\n\n"))
} else {
	suppressPackageStartupMessages(library("optparse"))
}

option_list = list(
	make_option(c("-b", "--ipBAM"), type="character", default=NULL, 
              help="BAM file for IP sample", metavar="character"),
	make_option(c("-i", "--inputBAM"), type="character", default=NULL, 
              help="BAM file for INPUT sample", metavar="character"),
	make_option(c("-m", "--method"), type="character", default=NULL, 
              help="Method used to identify modifications (MACS2, exomePeak, BayesPeak or MeTPeak)", metavar="character"),
	make_option(c("-r", "--gtf"), type="character", default=NULL, 
              help="GTF format of reference genome annotation", metavar="character"),	  
	make_option(c("-o", "--outfile"), type="character", default=NULL, 
              help="Output file", metavar="character")
); 
 
opt_parser = OptionParser(option_list=option_list);
opt = parse_args(opt_parser);

if(is.null(opt$method) | is.null(opt$ipBAM) | is.null(opt$inputBAM) | is.null(opt$gtf)) {
   message(paste0("\n\nRequired parameter inputs are missing!\n\n"))
   print_help(opt_parser)
   stop()
}

if(opt$method == "MACS2") {
	macs2.location <- Sys.which("macs2")
	if(macs2.location == ""){
		stop(paste0("\n\n", opt$method, " is not available\n\n"))
		print_help(opt_parser)
	}
} else if (opt$method == "exomePeak" | opt$method == "BayesPeak") {
	if (!opt$method %in% rownames(installed.packages())) {
		stop(paste0("\n\n", opt$method, " package is not available. \nPlease install it in R\n\n"))
	} else {
		suppressPackageStartupMessages(library(opt$method,character.only=TRUE))
	}
} else if (opt$method == "MeTPeak") {
	if (!opt$method %in% rownames(installed.packages())) {
		stop(paste0("\n\n", opt$method, " package not available.\nPlease install it (https://github.com/compgenomics/MeTPeak) in R\n\n"))
	} else {
		suppressPackageStartupMessages(library(opt$method,character.only=TRUE))
	}
	if (!'seqinr' %in% rownames(installed.packages())) {
		stop(paste0("\nseqinr package not available.\nPlease install it in R\n\n"))	
	} else {
		suppressPackageStartupMessages(library("seqinr"))
	}
}  else if(!is.null(opt$method)) {
    stop(paste0(opt$method, " is not supported!.\n\nOnly MACS2, exomePeak, BayesPeak or MeTPeak is supported!\n\n"))
	print_help(opt_parser)
} else {
	print_help(opt_parser)
	stop()
}

modRes <- ModCalling(method = opt$method, IPBAM = opt$ipBAM, inputBAM = opt$inputBAM, GTF = opt$gtf)
cat(paste0("Output to ",opt$outfile," ...", "\n"))
write.table(modRes, file = opt$outfile, sep = "\t", quote = F, row.names = F, col.names = F)
cat(paste0("Completed!", "\n"))

